function halt(msg, obj_) {
  console.error(msg, obj_);
  if (obj_) msg += "\n" + obj;
  throw msg;
}

function saturate(n) {
  return n < 0.0 ? 0.0 : n > 1.0 ? 1.0 : n;
}

function clamp(n, min, max) {
  return n < min ? min : n > max ? max : n;
}

function lerp(a, b, i) {
  return a + (b - a) * i;
}

function smoothstep(a, b, i) {
  i = saturate(i);
  return lerp(a, b, (3.0 - i * 2.0) * i * i);
}

function random(v_) {
  return v_
    ? v_.set(
      Math.random(),
      Math.random(),
      Math.random())
    : Math.random();
}

function randomRange(a, b, v_) {
  var d = b - a;
  return v_
    ? v_.set(
      Math.random() * d + a,
      Math.random() * d + a,
      Math.random() * d + a)
    : Math.random() * d + a;
}

function randomChoose(l) {
  return l[(Math.random() * l.length) ^ 0];
}

function property(obj, key, get_, set_) {
  var map = {};
  if (get_) map.get = get_;
  if (set_) map.set = set_;
  Object.defineProperty(obj, key, map);
}

function AnimationFrameInfo(totalTime_, elapsedTime_, frameNumber_) {
  var _totalTime = totalTime_ || 0;
  var _elapsedTime = elapsedTime_ || 0;
  var _frameNumber = frameNumber_ || 0;
  property(this, "totalTime", function () { return _totalTime * 0.001; });
  property(this, "totalTimeMilliseconds", function () { return _totalTime; });
  property(this, "elapsedTime", function () { return _elapsedTime * 0.001; });
  property(this, "elapsedTimeMilliseconds", function () { return _elapsedTime; });
  property(this, "frameNumber", function () { return _frameNumber; });
}

function AnimationTimer(paused_) {
  var _baseTime;
  var _lastElapsedTime;
  var _stopTime;
  var _paused;
  var _frameNumber;
  var _frameInfo;
  property(this, "frameInfo", function () { return _frameInfo; });
  property(this, "paused", function () { return _paused; });
  this.update = function (forAdvance_) {
    var now = performance.now();
    _frameInfo = new AnimationFrameInfo(now - _baseTime, now - _lastElapsedTime, _frameNumber);
    _frameNumber++;
    _lastElapsedTime = now;
  };
  this.reset = function () {
    var now = performance.now();
    _baseTime = now;
    _lastElapsedTime = now;
    _stopTime = now;
    _paused = true;
    _frameNumber = 0;
    _frameInfo = new AnimationFrameInfo();
  };
  this.restart = function () {
    this.reset();
    this.start();
  };
  this.pause = function () {
    if (!this._paused) {
      var now = performance.now();
      _stopTime = now;
      _paused = true;
    }
  };
  this.start = function () {
    if (this._paused) {
      var now = performance.now();
      var stopTime = now - _stopTime;
      _baseTime += stopTime;
      _lastElapsedTime += stopTime;
      _stopTime = 0;
      _paused = false;
    }
  };
  if (paused_) this.reset();
  else this.restart();
}

function RenderLoop() {
  var _requests = [];
  var _running = false;
  var _animationTimer = new AnimationTimer();
  function _requestAnimationFrame() {
    if (_requests.length) {
      requestAnimationFrame(function () {
        _animationTimer.update();
        for (var i in _requests)
          if (_requests[i].update)
            _requests[i].update(_animationTimer.frameInfo);
        for (var i in _requests)
          _requests[i].render(_animationTimer.frameInfo);
        for (var i = 0; i < _requests.length;) {
          if (_requests[i].once) {
            _requests.splice(i, 1);
            continue;
          }
          ++i;
        }
        _requestAnimationFrame();
      });
    } else {
      _running = false;
      console.info("[RenderLoop] End render loop.");
    }
  }
  function _launch() {
    if (!_running) {
      console.info("[RenderLoop] Start render loop.");
      _requestAnimationFrame();
      _running = true;
    }
  }
  this.request = function (render, update_) {
    _requests.push({ render: render, update: update_, once: false });
    _launch();
  };
  this.requestOnce = function (render, update_) {
    _requests.push({ render: render, update: update_, once: true });
    _launch();
  };
  this.discardAll = function () {
    _requests = [];
  };
}

function App() {
  var _renderer = new THREE.WebGLRenderer({ canvas: $("#Polycircuit")[0] });
  var _renderLoop = new RenderLoop();
  var _aspect;
  var _camera;
  
  _renderer.autoClearColor = false;

  var _circuitScene = new THREE.Scene();
  var _circuit;
  var _circuitMaterial = new THREE.ShaderMaterial({
    defines: {
      BLUR_STEP: 3,
      BLUR_SIZE: 0.015,
    },
    uniforms: {
      aspect: { type: "f", value: 0.0 },
      totalTime: { type: "f", value: 0.0 },
      elapsedTime: { type: "f", value: 0.0 },
      rotation0: { type: "f", value: 0.0 },
      rotation1: { type: "f", value: 0.0 },
      scaling0: { type: "f", value: 1.0 },
      scaling1: { type: "f", value: 1.0 },
      translation0: { type: "v2", value: new THREE.Vector2() },
      translation1: { type: "v2", value: new THREE.Vector2() },
      distortion0: { type: "v2", value: new THREE.Vector2() },
      distortion1: { type: "v2", value: new THREE.Vector2() },
    },
    vertexShader: [
      "varying vec2 vUv;",
      "varying vec3 vPosition;",
      "void main () {",
      "  vUv = uv;",
      "  vPosition = position;",
      "  gl_Position = projectionMatrix * modelViewMatrix * vec4(position, 1);",
      "}"].join("\n"),
    fragmentShader: [
      "varying vec2 vUv;",
      "uniform float aspect, totalTime, elapsedTime, rotation0, rotation1, scaling0, scaling1;",
      "uniform vec2 translation0, translation1, distortion0, distortion1;",
      "vec2 rotate (in vec2 p, float rotation) {",
      "  float s = sin(rotation), c = cos(rotation);",
      "  return vec2(p.x * c - p.y * s, p.x * s + p.y * c);",
      "}",

      //
      // Following part based on this post:
      // http://glsl.heroku.com/e#12858.0
      // by srtuss, 2013
      //

      "vec2 random (in vec2 p, in vec2 d) {",
      "  return fract(vec2(",
      "    sin(p.x * 48.532 + p.y * 374.002 + d.x),",
      "    cos(p.x * 125.57 + p.y * 45.763 + d.y)));",
      "}",
      "float voronoi (in vec2 p, in vec2 d, int k) {",
      "  vec2 a = floor(p), b = fract(p), c = vec2(8);",
      "  for (int y = -1; y <= 1; ++y) {",
      "    for (int x = -1; x <= 1; ++x) {",
      "      vec2 q = vec2(x, y), t = abs(q + random(a + q, d) - b);",
      "      float m = max(t.x, t.y);",
      "      if (m < c.x) {",
      "        c = vec2(m, c.x);",
      "      } else if (m < c.y) {",
      "        c = vec2(k, m);",
      "      }",
      "    }",
      "  }",
      "  return c.y - c.x;",
      "}",

      "vec3 color (float r, float s, in vec2 t, in vec2 d) {",
      "  vec2 p = rotate((vUv * 2.0 - 1.0) * vec2(aspect, 1.0) * s, r) + t;",
      "  int h = ",
      "    int(voronoi(p * 1.27 + 0.11, d, 0) < 0.02) +",
      "    int(voronoi(p * 1.85 + 0.25, d, 1) < 0.05) -",
      "    int(voronoi(p * 2.53 + 0.13, d, 0) < 0.02) -",
      "    int(voronoi(p * 2.82 + 0.55, d, -1) < 0.08);",
      "  vec3 c;",
      "  if (h > 1) {",
      "    c = vec3(1, 0.95, 0.7);",
      "  } else if (h > 0) {",
      "    c = vec3(1, 0, 0.5);",
      "  } else if (h > -1) {",
      "    c = vec3(0.03, 0.06, 0.09);",
      "  } else {",
      "    c = vec3(0, 0.2, 0.3);",
      "  }",
      "  return c;",
      "}",
      "void main() {",
      "  vec3 f = vec3(0);",
      "  for (int i = -BLUR_STEP; i <= BLUR_STEP; ++i) {",
      "    float a = BLUR_SIZE / elapsedTime * float(i) / float(BLUR_STEP);",
      "    float r = rotation0 + (rotation0 - rotation1) * a;",
      "    float s = scaling0 + (scaling0 - scaling1) * a;",
      "    vec2 t = translation0 + (translation0 - translation1) * a;",
      "    vec2 d = distortion0 + (distortion0 - distortion1) * a;",
      "    f += color(r, s, t, d) * abs(1.0 - float(i) / float(BLUR_STEP + 1));", // TODO
      "  }",
      "  gl_FragColor = vec4(f / float(BLUR_STEP * 2 + 1), 1);",
      "}"].join("\n"),
  });

  var _postProcessScene = new THREE.Scene();
  var _postProcess;
  var _postProcessColor;
  var _postProcessMaterial = new THREE.ShaderMaterial({
    uniforms: {
      color: { type: "t", value: null },
      totalTime: { type: "f", value: 0.0 },
      jittering: { type: "f", value: 0.0 },
    },
    vertexShader: [
      "varying vec2 vUv;",
      "varying vec3 vPosition;",
      "void main () {",
      "  vUv = uv;",
      "  vPosition = position;",
      "  gl_Position = projectionMatrix * modelViewMatrix * vec4(position, 1);",
      "}"].join("\n"),
    fragmentShader: [
      "varying vec2 vUv;",
      "uniform sampler2D color;",
      "uniform float totalTime, jittering;",
      "void main() {",
      "  float w = pow(jittering, 0.5);",
      "  float k = cos(w * 123.29 + totalTime) * w;",
      "  vec2 p = (vUv + vec2(k, w * 0.25)) * vec2(1.0 - w, 1);",
      "  vec2 d = vec2(0.001 + k * 0.005, 0);",
      "  vec3 c = vec3(",
      "    texture2D(color, p - d).r,",
      "    texture2D(color, p).g,",
      "    texture2D(color, p + d).b);",
      "  vec2 j = vUv * 2.0 - 1.0;",
      "  c = mix(c, c + max(dot(c, vec3(0.25)), 0.0), w) *",
      "    (1.2 - pow(length(j * 0.75), 2.0));",
      "  gl_FragColor = vec4(c, 1.0);",
      "}"].join("\n"),
  });

  function _scheduleDistortionXTween() {
    new TWEEN.Tween(_circuitMaterial.uniforms.distortion0.value)
      .to({ x: _circuitMaterial.uniforms.distortion0.value.x + randomChoose([-4.0, -2.0, 2.0, 4.0]) }, 3000)
      .delay(randomChoose([0, 250, 500, 2500, 3500, 4500, 5000]))
      .easing(TWEEN.Easing.Quadratic.InOut)
      .onComplete(_scheduleDistortionXTween)
      .start();
  }
  function _scheduleDistortionYTween() {
    new TWEEN.Tween(_circuitMaterial.uniforms.distortion0.value)
      .to({ y: _circuitMaterial.uniforms.distortion0.value.y + randomChoose([-4.0, -2.0, 2.0, 4.0]) }, 3000)
      .delay(randomChoose([0, 250, 500, 2500, 3500, 4500, 5000]))
      .easing(TWEEN.Easing.Quadratic.InOut)
      .onComplete(_scheduleDistortionYTween)
      .start();
  }
  function _scheduleRotationTween() {
    new TWEEN.Tween(_circuitMaterial.uniforms.rotation0)
      .to({ value: _circuitMaterial.uniforms.rotation0.value + randomChoose([-0.5, -0.25, 0.25, 0.5]) * Math.PI }, 250)
      .delay(randomChoose([0, 0, 500, 1500, 3000, 5000]))
      .easing(TWEEN.Easing.Back.Out)
      .onComplete(_scheduleRotationTween)
      .start();
  }
  function _scheduleScalingTween() {
    new TWEEN.Tween(_circuitMaterial.uniforms.scaling0)
      .to({ value: randomChoose([0.25, 0.5, 0.5, 1.0]) }, randomChoose([250, 500, 500, 1000]))
      .delay(randomChoose([0, 0, 500, 1000, 1500]))
      .easing(TWEEN.Easing.Quadratic.Out)
      .onComplete(_scheduleScalingTween)
      .start();
  }
  function _scheduleTranslationTween() {
    var offsetX = randomChoose([-4.0, -3.0, -2.0, 0.0, 0.0, 0.0, 0.0, 2.0, 3.0, 4.0]);
    var offsetY = offsetX * randomChoose([-1.0, 0.0, 1.0]);
    new TWEEN.Tween(_circuitMaterial.uniforms.translation0.value)
      .to({
        x: _circuitMaterial.uniforms.translation0.value.x + offsetX,
        y: _circuitMaterial.uniforms.translation0.value.y + offsetY,
      }, 2500)
      .onComplete(_scheduleTranslationTween)
      .start();
  }
  function _scheduleJitteringTween() {
    new TWEEN.Tween(_postProcessMaterial.uniforms.jittering)
      .to({ value: 1.0 }, randomChoose([100, 250, 500]))
      .delay(randomChoose([0, 500, 1000, 1500, 5000]))
      .onComplete(function () {
        _postProcessMaterial.uniforms.jittering.value = 0.0;
        _scheduleJitteringTween();
      })
      .start();
  }
  function _resize() {
    var width = $(window).width();
    var height = $(window).height();
    _renderer.setSize(width, height);
    _aspect = width / height;
    _camera = new THREE.OrthographicCamera(width / -2, width / 2, height / 2, height / -2, -10000, 10000);

    if (_circuit) _circuitScene.remove(_circuit);
    _circuit = new THREE.Mesh(new THREE.PlaneGeometry(width, height), _circuitMaterial);
    _circuitScene.add(_circuit);

    if (_postProcess) _postProcessScene.remove(_postProcess);
    _postProcess = new THREE.Mesh(new THREE.PlaneGeometry(width, height), _postProcessMaterial);
    _postProcessScene.add(_postProcess);
    _postProcessColor = new THREE.WebGLRenderTarget(width, height, {
      minFilter: THREE.PointFilter,
      magFilter: THREE.PointFilter,
    });
  }
  function _update(frameTime) {
    _circuitMaterial.uniforms.totalTime.value = frameTime.totalTime;
    _circuitMaterial.uniforms.elapsedTime.value = frameTime.elapsedTime;
    _circuitMaterial.uniforms.rotation1.value = _circuitMaterial.uniforms.rotation0.value;
    _circuitMaterial.uniforms.scaling1.value = _circuitMaterial.uniforms.scaling0.value;
    _circuitMaterial.uniforms.distortion1.value.x = _circuitMaterial.uniforms.distortion0.value.x;
    _circuitMaterial.uniforms.distortion1.value.y = _circuitMaterial.uniforms.distortion0.value.y;
    _circuitMaterial.uniforms.translation1.value.x = _circuitMaterial.uniforms.translation0.value.x;
    _circuitMaterial.uniforms.translation1.value.y = _circuitMaterial.uniforms.translation0.value.y;
    _circuitMaterial.uniforms.aspect.value = _aspect;
    _postProcessMaterial.uniforms.color.value = _postProcessColor;
    _postProcessMaterial.uniforms.totalTime.value = frameTime.totalTime;
    TWEEN.update(frameTime.totalTimeMilliseconds);
  }
  function _render(frameTime) {
    _renderer.render(_circuitScene, _camera, _postProcessColor, false);
    _renderer.render(_postProcessScene, _camera);
  }

  _scheduleDistortionXTween();
  _scheduleDistortionYTween();
  _scheduleRotationTween();
  _scheduleScalingTween();
  _scheduleTranslationTween();
  _scheduleJitteringTween();
  _resize(); $(window).resize(_resize);
  _renderLoop.request(_update, _render);
}

var app;

$(document).ready(function () {
  app = new App();
});
